;                           WTEST.ASM                    Agner Fog 2004-01-22
comment #
This is a general program for testing how many clock cycles a piece of
code takes on a microprocessor.
 
Your piece of code is repeated several times and the number of clock cycles
is counted each time.

The code is run under Windows in 32-bit console mode. 

It is not possible to turn off interrupts in protected mode in Windows 2000 
and later versions of Windows. Instead, the test code is executed with a 
high priority to minimize the number of interfering interrupts. The code may
sometimes take longer time than expected because of interrupts. The smallest
clock count is usually the true time consumption when code and data are cached
and branch predictors have been initialized. The first few clock counts are 
always longer because code and data are not cached.

The test program will turn off interrupts when run under Windows 95 and 
Windows 98. This will make the clock counts more accurate for code that takes
so long time that it is likely to be interrupted if interrupts are enabled.

Instructions:
-------------

User-definable areas are marked with stars ***.

Insert the piece of code you want to test after the line marked with:
"***   BEGIN INSTRUCTIONS TO TEST   ***"
Any necessary data or initializations may be inserted in the appropriately
marked spaces.

It is not recommended to output any data to the screen from the piece of 
code in the INSTRUCTIONS TO TEST field. If you want to output any results
from your code then use CALL UOUT. This procedure stores the contents of the
EAX register to be output after the testing is finished.

This test program is written for MASM version 6.x.
Use WT.BAT to assemble and link with KERNEL32.LIB.

 2004 GNU General Public License www.gnu.org/copyleft/gpl.html
#

;***********************   SET OPTIONS   ***********************

; set ITER = desired number of iterations
ITER = 30


.LISTALL

; =========================   Microprocessor version   =========================

.686
.model flat
.mmx
.xmm

; =======================   system function prototypes   =======================

ExitProcess PROTO NEAR32 stdcall,
    dwExitCode:DWORD

GetStdHandle PROTO NEAR32 stdcall,
    nStdHandle:DWORD
STD_OUTPUT  EQU -11

WriteFile PROTO NEAR32 stdcall,
    hFile:DWORD, lpBuffer:NEAR32, nNumberOfCharsToWrite:DWORD,
    lpNumberOfBytesWritten:NEAR32, lpOverlapped:NEAR32

GetCurrentProcess PROTO NEAR32 stdcall

SetPriorityClass PROTO NEAR32 stdcall,
    hProcess:DWORD, dwPriorityClass:DWORD
NORMAL_PRIORITY_CLASS   EQU    20h
HIGH_PRIORITY_CLASS     EQU    80h
REALTIME_PRIORITY_CLASS	EQU   100h

GetVersionExA PROTO NEAR32 stdcall,
    lpVersionInformation:NEAR32

; =======================   SEGMENT DEFINITIONS   =======================

_TEXT   SEGMENT USE32 PARA PUBLIC 'CODE'	
_TEXT   ENDS

; ====================   STACK SEGMENT   ====================

STACK   SEGMENT USE32 PARA STACK 'STACK'
        DD      400H DUP (?)
STACK   ENDS

; ====================   DATA SEGMENT   ====================

_DATA1  SEGMENT USE32 PAGE PUBLIC 'DATA'	

; text strings
Heading         BYTE    13,10,'Clocks      ',0
UHead           BYTE    '        Additional user data:',13,10,0
NoTSC           BYTE    'This microprocessor has no time stamp counter',13,10,0
NewLine         BYTE    13,10,0
Space2          BYTE    '  ',0

ALIGN 2
; List of timing overhead for different microprocessors.
; You may adjust these values so that a clock count of 0 is shown when 
; the INSTRUCTIONS TO TEST field is empty.
; ** Missing values may be inserted **
OVERHEADS       LABEL WORD             ; table entry. don't change order
UNIDENTIFIED    WORD     0             ; processor has no CPUID
INTEL13         WORD     0,0,0         ; family 1 - 3, obsolete
INTEL4          WORD     0             ; 80486
INTEL5          WORD    46             ; Pentium 1, Pentium MMX
INTEL6          WORD   162             ; Pentium Pro, Pentium II, Pentium III
INTEL7          WORD   636             ; Pentium 4
AMD4            WORD   192 ; ?         ; Am486, Am5x86, K5
AMD5            WORD     0             ; K6
AMD6            WORD     0             ; Athlon, Duron
AMD7            WORD     0             ; Athlon64, Opteron

; test data
ALIGN 8
StartTime       QWORD   ?              ; clock cycles start count
TimeOverhead    DWORD   ?              ; entry picked from above list
counter         DWORD   0              ; loop counter
espsave         DWORD   ?              ; save stack pointer
processortype   DWORD   ?              ; result of DetectProcessor call
ProcessHandle   DWORD   ?              ; handle of current process
ClearInterrupt  DWORD   ?              ; clear interrupt allowed

ALIGN 8
; storage for output data:
clocklist       QWORD   ITER dup (?)   ; clock cycles
userlist        DWORD   100  dup (?)   ; any user data
nuser           DWORD   0              ; number of data from UOUT calls

align 16
;*******************   USER DATA BEGIN   ******************************
; put any data here that your test code needs

D0      DWORD   0
D1      DWORD   0
D2      DWORD   0
D3      DWORD   0
D4      DWORD   0
D5      DWORD   0
        DWORD   4096 dup (0)
TestString      BYTE '8456784567845678123',0

; ====================   USER DATA END   ====================

_DATA1  ENDS

; ====================   CODE SEGMENT   ====================

_TEXT   SEGMENT USE32 PARA PUBLIC 'CODE'	
ASSUME  CS:FLAT, DS:FLAT
_start:
PUBLIC _start                          ; make entry point public

        ; various initializations
        call    DetectProcessor        ; detect processor features
        mov     [processortype], eax
        test    eax, 1 SHL 13          ; test if time stamp counter supported
        jnz     S10
        push    offset NoTSC           ; RDTSC not supported by microprocessor
        call    WriteString            ; write error message
        jmp     S20                    ; exit
S10:    push    offset Heading         ; write heading
        call    WriteString

        ; get timing overhead from table
        mov     eax, [processortype]
        mov     ebx, eax
        shr     eax, 4
        and     eax, 7                 ; microprocessor family
        shr     ebx, 6
        and     ebx, 8                 ; AMD processor
        or      eax, ebx               ; bit 0-2 = family, bit 3 = vendor
        movzx   eax, overheads[eax*2]  ; find timing overhead in table
        mov     [TimeOverhead], eax

        ; detect if CLI allowed by operating system
; define which versions of Windows allow CLI:
Min_OS_CLI_allowed = 04000000H         ; first Windows version that allows CLI = Win95
Max_OS_CLI_allowed = 040A08FFH         ; last  Windows version that allows CLI = Win98SE

        call    OSVersion              ; check operating system version
        xor     ebx, ebx
        cmp     eax, Min_OS_CLI_allowed
        jb      CLI_not_allowed
        cmp     eax, Max_OS_CLI_allowed
        ja      CLI_not_allowed
        inc     ebx
CLI_not_allowed:
        mov     [ClearInterrupt], ebx

        ; set maximum priority to minimize interfering task switches:
        INVOKE  GetCurrentProcess
        mov     [ProcessHandle], eax
        INVOKE  SetPriorityClass , eax, REALTIME_PRIORITY_CLASS
        
        ; initializations of microprocessor
        finit                          ; initialize floating point registers
        cld                            ; clear direction flag        
        test    [processortype], 1 SHL 11  ; can we use XMM registers?
        jz      NO_MMX
I = 0   ; reset all XMM registers:
REPT 8  ; repeat for all 8 XMM registers:
        xorps   xmm(I),xmm(I)          ; set to 0
I = I + 1
ENDM
NO_MMX: mov     [espsave], esp         ; save stack pointer
        cmp     [ClearInterrupt], 0    ; clear interrupt if possible
        je      NoClear
        CLI                            ; CLI allowed in Win95 and Win98 only
NoClear:
 
;*******************   BEGIN USER INITIALIZATIONS 1   ***********************
; insert any initializations here that you want to run only once:

        xor    edi, edi

        
;=======================   END USER INITIALIZATIONS 1   =======================
             
ALIGN   16
AGAIN:  ; top of loop. repeat test code ITER times

;****************   BEGIN USER INITIALIZATIONS 2 ****************************
; insert any initializations here that you want to run for each iteration:
; (only ESI and EDI are unchanged between iterations)

        mov esi, offset D0


;=======================   END USER INITIALIZATIONS 2   =======================

        ; start to count clock cycles:
        xor     eax, eax
        cpuid                          ; serialize
        rdtsc                          ; read clock counter
        mov     dword ptr [StartTime], eax
        mov     dword ptr [StartTime+4], edx
        xor     eax, eax
        cpuid                          ; serialize again

; *********************   BEGIN INSTRUCTIONS TO TEST   **********************
; insert the instructions to test here:



  ; ----------------------------------------------------------
  ; Convert positive decimal ASCII string into QWORD value
  ; by IanB

  ; On entry:
  ;     string pointer in ECX, string must be zero-terminated
  ; On exit:
  ;     converted value in EAX (low DWORD)/ EDX (high DWORD)
  ;     ECX is zero if there is an error in string or overflow,
  ;     else non-zero to show successful conversion
  ; EBX/EBP/EDI/ESI preserved
  ; ----------------------------------------------------------

        xor     eax, eax            ; use EAX as low-DWORD accumulator
        xor     edx, edx            ; use EDX as high-DWORD accumulator
        lea     ecx, TestString
        push    ebx                 ; use EBX as new byte input
        push    esi                 ; use these as convenience storage
        push    edi

        mov     al, BYTE PTR [ecx]
        add     ecx, 1
        xor     ebx, ebx
        sub     al, 30H
        jb      BadNumericString    ; not numeric digit
        cmp     al, 9
        ja      BadNumericString    ; not numeric digit

NewASCIIDigitLoop:
        mov     bl, BYTE PTR [ecx]
        add     ecx, 1
        mov     esi, 9              ; add all possible multiplication carries here
        xor     edi, edi

        test    bl, bl
        jz      EndASCIIDigitLoop   ; found string end
        sub     bl, 30H
        jb      BadNumericString    ; not numeric digit
        cmp     bl, 9
        ja      BadNumericString    ; not numeric digit
        cmp     edx, 19999999H
        ja      BadNumericString    ; value too big for QWORD when multiplied by 10
        je      CriticalTest        ; test low DWORD if reached critical value

MultiplyQWBy10:
        add     eax, eax            ; current value * 2
        jc      @F
        sub     esi, 5              ; remove carries for this multiplication
@@:
        add     edx, edx
        add     edi, eax            ; save value * 2
        add     esi, edx
        add     eax, eax            ; current value * 4
        jc      @F
        sub     esi, 2              ; remove carries for this multiplication
@@:
        add     edx, edx
        add     eax, eax            ; current value * 8
        jc      @F
        sub     esi, 1              ; remove carries for this multiplication
@@:
        add     edx, edx
        add     eax, edi            ; current value * 10 = 8 + 2
        jc      @F
        sub     esi, 1              ; remove final pre-added carry
@@:
        add     edx, esi
        add     eax, ebx            ; add new digit
        jc      @F
        xor     ebx, ebx
        jmp     NewASCIIDigitLoop
@@:
        add     edx, 1
        jc      BadNumericString    ; value too big for QWORD
        xor     ebx, ebx
        jmp     NewASCIIDigitLoop

CriticalTest:
        cmp     eax, 99999999H      ; did we reach the critical value?
        jbe     MultiplyQWBy10

BadNumericString:
        xor     eax, eax
        xor     edx, edx
        xor     ecx, ecx            ; show non-successful result by zero in ECX
        jmp     @F

EndASCIIDigitLoop:
        or      ecx, 0ffffffffH     ; show successful result by non-zero in ECX
@@:
        pop     edi
        pop     esi
        pop     ebx




;=======================    END INSTRUCTIONS TO TEST    =======================

        xor     eax, eax
        cpuid                          ; serialize
        rdtsc                          ; read clock
        mov     ebp, [counter]
        mov     dword ptr clocklist[ebp], eax    ; store clock count in table
        mov     dword ptr clocklist[ebp+4], edx
        xor     eax, eax
        cpuid                          ; serialize again
        mov     eax, dword ptr [StartTime]
        mov     edx, dword ptr [StartTime+4]
        add     eax, [TimeOverhead]    ; subtract the number of clocks used by surrounding code
        adc     edx, 0
        sub     dword ptr clocklist[ebp], eax    ; subtract overhead from clock count
        sbb     dword ptr clocklist[ebp+4], edx
        ; repetion loop ends here
        mov     esp, [espsave]         ; restore ESP in case it has been changed
        add     ebp, TYPE clocklist    ; increment clocklist index
        mov     [counter], ebp
        cmp     ebp, SIZE clocklist
        jb      AGAIN                  ; repeat test ITER times

; ====================   TEST FINISHED. WRITE RESULTS   ====================

        ; set interrupt if cleared
        cmp     [ClearInterrupt], 0
        je      NoSTI
        STI                            ; set interrupt if it was cleared
NoSTI:
        ; set process priority back to normal
        INVOKE  SetPriorityClass , ProcessHandle, NORMAL_PRIORITY_CLASS
        
        ; loop through result lists
        xor     edi, edi        
NEXTRESULT:
        push    offset NewLine         ; new line
        call    WriteString
        lea     eax, clocklist[edi]    ; point to clock count
        mov     ecx, 10                ; string length
        call    WriteDec64             ; write clock count
        add     edi, TYPE clocklist    ; increment clocklist index
        cmp     edi, SIZE clocklist
        jb      NEXTRESULT             ; loop through clocklist
        
        ; write any data saved with call UOUT
        xor     esi, esi
        cmp     esi, [nuser]
        je      S20
        push    offset UHead           ; write heading for user data list
        call    WriteString
S18:    cmp     esi, [nuser]           ; loop through user list
        jnb     S19
        mov     eax, userlist[esi*4]
        call    WriteHex               ; write as hexadecimal
        push    offset Space2          ; write spaces
        call    WriteString
        inc     esi
        jmp     S18                    ; loop
S19:    
S20:    ; clean up and finish
        finit                          ; clear floating point registers
        INVOKE  ExitProcess, 0         ; exit with return code 0

        
; ====================   SUBROUTINES   ====================

; user procedure for delayed output of additional data EAX
UOUT    PROC    NEAR
        push    ebx
        mov     ebx, [nuser]
        cmp     ebx, LENGTH userlist   ; check if list full
        jnb     UO9
        mov     userlist[ebx*4], eax   ; store in userlist
        inc     ebx
        mov     [nuser], ebx           ; number of data stored
UO9:    pop     ebx
        ret
UOUT    ENDP


; Write zero-terminated text string (stdcall)
WriteString PROC NEAR
        push   ebx
        INVOKE GetStdHandle, STD_OUTPUT  ; get handle for console output
        mov    edx, [esp+8]            ; string pointer
        ; find string length
        mov    ecx, edx
Search0:cmp    BYTE PTR [ecx], 0
        je     EndS0
        inc    ecx
        jmp    Search0
EndS0:  sub    ecx, edx                ; length = end - begin
        push   0                       ; store bytes written here
        mov    ebx, esp                ; pointer to bytes written
        INVOKE WriteFile,
        eax,                           ; file handle for console out
        edx,                           ; address of string
        ecx,                           ; length of string
        ebx,                           ; pointer to bytes written
        0                              ; overlapped mode
        pop    eax                     ; bytes written
        pop    ebx                     ; restore ebx
        ret    4
WriteString ENDP


; write EAX as hexadecimal number with 8 digits
WriteHex PROC NEAR
        push    ebx
        sub     esp, 20                ; local storage
        mov     edx, esp               ; pointer to text buffer
        mov     ecx, 8                 ; loop counter
wh1:    rol     eax, 4                 ; get most significant bits first
        mov     ebx, eax
        and     ebx, 0FH               ; isolate 4 bits
        add     ebx, '0'               ; convert to ASCII
        cmp     ebx, '9'
        jna     wh2
        add     ebx, 'A'-'9'-1
wh2:    mov     [edx], bl              ; store ASCII in buffer
        inc     edx                    ; increment pointer
        loop    wh1                    ; repeat 8 times
        mov     byte ptr [edx], 0      ; terminate string
        push    esp                    ; pointer to buffer
        call    WriteString            ; write text        
        add     esp, 20                ; release local buffer
        pop     ebx
        ret
WriteHex ENDP


; Write a decimal signed integer, right justified. Input:
; EAX: number to write
; ECX: desired length of string (max 80)
; Leading spaces will be added if the number uses less than ECX characters
WriteDec PROC NEAR
        push    ebx
        push    esi
        push    edi
        lea     edi, [esp-4]           ; point to end of string buffer
        sub     esp, 88                ; make space for string buffer
        cmp     ecx, 80
        jna     WD1
        mov     ecx, 20                ; error: ECX too big
WD1:    mov     esi, eax               ; remember sign
        cdq                            ; make EAX positive
        xor     eax, edx
        sub     eax, edx               ; = absolute value
        mov     byte ptr [edi], 0      ; terminate string with 0
        dec     edi                    ; string is written backwards
        mov     ebx, 10                ; divisor for conversion to decimal
WD2:    xor     edx, edx               ; division loop starts here
        div     ebx                    ; divide by 10
        add     edx, '0'               ; remainder. convert to ASCII
        mov     [edi], dl              ; write digit
        dec     edi                    ; decrement string pointer
        dec     ecx                    ; count down string length
        test    eax, eax
        jnz     WD2                    ; loop while number not 0
        test    esi, esi               ; test sign
        jns     WD3
        mov     byte ptr [edi],'-'     ; write minus sign
        dec     edi
        dec     ecx
WD3:    test    ecx, ecx               ; write leading spaces while ECX > 0
        jle     WD4
        mov     byte ptr [edi],' '     ; write space
        dec     edi
        dec     ecx
        jmp     WD3
WD4:    inc     edi
        push    edi                    ; pointer to string buffer
        call    WriteString            ; write text        
        add     esp, 88                ; release local space
        pop     edi
        pop     esi
        pop     ebx
        ret
WriteDec      ENDP


; Write a 64-bit signed integer as decimal, right justified. Input:
; EAX: pointer to 64-bit integer to write (max 10^18)
; ECX: desired length of string (max 80)
; Leading spaces will be added if the number uses less than ECX characters
WriteDec64 PROC NEAR
WDMAP   STRUCT                         ; local data on stack
textb1  BYTE    104 dup (?)            ; text buffer for leading spaces
textb2  BYTE    1   dup (?)            ; text buffer for sign
textb3  BYTE    23  dup (?)            ; text buffer for number
bcdnum  TBYTE   ?                      ; BCD number
        WORD    ?                      ; align by 4
WDMAP   ENDS

        push    ebx
        push    esi
        push    edi
        sub     esp, size WDMAP        ; make space for local data
        fild    qword ptr [eax]        ; load 64-bit integer
        fabs                           ; remove sign
        fbstp   [esp].WDMAP.bcdnum     ; convert to BCD
        mov     ebx, [eax+4]           ; save sign
        cmp     ecx, 80                ; check if ECX too high
        jna     WD10
        mov     ecx, 20
WD10:   lea     esi, [esp].WDMAP.bcdnum+9 ; point to high end of BCD number
        lea     edi, [esp].WDMAP.textb3   ; point to text buffer
        xor     edx, edx               ; count digits written in DL
WD11:   ; loop through BCD number, most significant digit first
        mov     al, [esi]              ; two packed digits
        mov     ah, al
        shr     ah, 4                  ; first digit
        and     al, 0FH                ; second digit
        mov     dh, ah                 ; test first digit
        or      dh, dl
        jz      WD12                   ; skip leading zero
        add     ah, '0'                ; convert to ASCII
        mov     [edi], ah              ; store in string
        inc     edi                    ; increment string pointer
        inc     edx                    ; count digit written
WD12:   mov     dh, al                 ; test second digit
        or      dh, dl
        jz      WD13                   ; skip leading zero
        add     al, '0'                ; convert to ASCII
        mov     [edi], al              ; store in string
        inc     edi
        inc     edx                    ; count digit written
WD13:   dec     esi                    ; loop backwards through BCD number
        lea     eax, [esp].WDMAP.bcdnum ; point to low end of BCD number
        cmp     esi, eax
        jae     WD11                   ; loop
        test    dl, dl                 ; have we written any digits?
        jnz     WD14
        mov     byte ptr [edi], '0'    ; no digits written. Write a zero
        inc     edi
        inc     edx
WD14:   mov     byte ptr [edi], 0      ; terminate string
        lea     esi, [esp].WDMAP.textb2   ; point to sign
        test    ebx, ebx               ; test if negative
        jns     WD15
        mov     byte ptr [esi], '-'    ; write minus
        dec     esi
        inc     edx                    ; count character written
WD15:   movzx   edx, dl
        sub     ecx, edx               ; desired length minus written length
WD16:   jle     WD17                   ; skip if not positive
        mov     byte ptr [esi], ' '    ; write spaces
        dec     esi
        sub     ecx, 1
        jmp     WD16                   ; loop while ECX not negative
WD17:   inc     esi                    ; point to begin of string
        push    esi                    ; pass pointer to WriteString
        call    WriteString            ; write text        
        add     esp, size WDMAP        ; release local space
        pop     edi
        pop     esi
        pop     ebx
        ret
WriteDec64      ENDP


; Detect microprocessor features
; return value:
; The return value is 0 if microprocessor has no CPUID instruction
; bits 0-3 = model
; bits 4-7 = family:  4 for 80486, Am486, Am5x86
;                     5 for P1, PMMX, K6
;                     6 for PPro, P2, P3, Athlon, Duron
;                    15 for P4, Athlon64, Opteron
; bit    8 = vendor is Intel
; bit    9 = vendor is AMD
; bit   11 = XMM registers enabled by operating system
; bit   12 = floating point instructions supported
; bit   13 = time stamp counter supported
; bit   14 = CMPXCHG8 instruction supported
; bit   15 = conditional move and FCOMI supported (PPro, P2, P3, P4, Athlon, Duron, Opteron)
; bit   23 = MMX instructions supported (PMMX, P2, P3, P4, K6, Athlon, Duron, Opteron)
; bit   24 = FXSAVE / FXRSTOR supported
; bit   25 = SSE instructions supported (P3, P4, Athlon64, Opteron)
; bit   26 = SSE2 instructions supported (P4, Athlon64, Opteron)
; bit   27 = SSE3 instructions supported (forthcoming "Prescott")
; bit   28 = hyperthreading (P4)
; AMD-specific instructions:
; bit   29 = 3DNow instructions
; bit   30 = 3DNow extension instructions
; bit   31 = MMX extension

DetectProcessor PROC NEAR
        push    ebx
        push    esi
        push    edi
        xor     edi, edi               ; will contain result
; detect if CPUID instruction supported by microprocessor:
        pushfd
        pop     eax
        xor     eax, 1 SHL 21          ; check if CPUID bit can toggle
        push    eax
        popfd
        pushfd
        pop     ebx
        xor     eax, ebx
        and     eax, 1 SHL 21
        jnz     DPEND                  ; CPUID not supported
        xor     eax, eax
        cpuid                          ; get number of CPUID functions
        test    eax, eax
        jz      DPEND                  ; function 1 not supported
        ; determine vendor
        xor     ebx, ebx
        cmp     ch, 't'
        sete    bl                     ; 1 if vendor = Intel
        cmp     ch, 'A'
        sete    dl   
        shl     dl, 1                  ; 2 if vendor = AMD
        or      bl, dl
        shl     ebx, 8                 ; shift vendor info to bit 8-9
        or      edi, ebx               ; EDI now contains vendor information
        mov     eax, 1
        cpuid                          ; get family and features
        shr     eax, 4
        and     eax, 0FFH              ; family and model
        or      edi, eax               ; combine bits in EDI
        and     ecx, 1                 ; SSE3 feature
        shl     ecx, 27                ; move to bit 27
        or      edi, ecx               ; combine bits in EDI
        mov     ecx, edx               ; feature flags
        and     ecx, 17FF8000H         ; these feature flags remain in same position
        or      edi, ecx               ; combine bits in EDI
        mov     ecx, edx               ; shuffle feature bits 0, 4, 8 to 12, 13, 14
        and     ecx, 1                 ; FPU feature
        shl     ecx, 12-0              ; move to bit 12
        or      edi, ecx               ; combine bits in EDI
        mov     ecx, edx
        and     ecx, 10H               ; time stamp counter feature
        shl     ecx, 13-4              ; move to bit 13
        or      edi, ecx               ; combine bits in EDI        
        mov     ecx, edx
        and     ecx, 100H              ; CMPXCHG8 feature
        shl     ecx, 14-8              ; move to bit 14
        or      edi, ecx               ; combine bits in EDI
        test    edi, 200H              ; is it an AMD?
        jz      Not_AMD
        ; Get AMD-specific features
        mov     eax, 80000000H
        cpuid
        cmp     EAX, 80000001H
        jb      Not_AMD
        mov     eax, 80000001H
        cpuid                          ; get AMD-specific features into EDX
        mov     eax, edx
        and     eax, 0C0000000H        ; 3DNow instructions support bit 30 - 31
        or      edi, eax               ; combine bits in EDI
        and     edx, 400000H           ; MMX extensions bit 22        
        shl     edx, 29-22             ; move to bit 29
        or      edi, edx               ; combine bits in EDI
Not_AMD:
        ; test OS support for XMM registers
        test    edi, 2000000H          ; SSE support
        jz      no_SSE_test
        test    edi, 1000000H          ; FXSAVE/FXRSTOR support
        jz      no_SSE_test
        smsw    ax                     ; read lower part of CR0 without violating privilege
        test    al, 4                  ; test if FXSAVE is emulated
        jnz     no_SSE_test        
        mov     esi, esp               ; save stack pointer
        sub     esp, 200H              ; allocate space for FXSAVE
        and     esp, -10H              ; align by 16
TESTDATA = 0D95A34BEH                  ; random test value
TESTPS   = 10CH                        ; position to write TESTDATA = upper part of XMM6 image
        fxsave  [esp]                  ; save FP/MMX and XMM registers
        mov     eax,[esp+TESTPS]       ; read part of XMM6 register
        xor     DWORD PTR [esp+TESTPS],TESTDATA  ; change value
        fxrstor [esp]                  ; load changed value into XMM6
        mov     [esp+TESTPS],eax       ; restore old value in buffer
        fxsave  [esp]                  ; save again
        mov     ebx,[esp+TESTPS]       ; read changed XMM6 register
        mov     [esp+TESTPS],eax       ; restore old value
        fxrstor [esp]                  ; load restored value into XMM6
        xor     eax, ebx               ; get difference between old and new value
        cmp     eax, TESTDATA          ; test if XMM6 was changed correctly
        sete    al                     ; 1 if test successful
        movzx   eax, al                ; clear rest of EAX
        shl     eax, 11                ; move to bit 11
        or      edi, eax               ; combine bits in EDI        
        mov     esp, esi               ; restore ESP
no_SSE_test:
DPEND:  
        mov     eax, edi               ; return result bits
        pop     edi
        pop     esi
        pop     ebx
        ret
DetectProcessor ENDP


; Get operating system version
; Return value in EAX:
; bit 24-31 = major version (Win95-98 = 4, Win2000/XP = 5)
; bit 16-23 = minor version (Win95 = 0, Win98 > 0)
; bit 0 -15 = build number (Win98 = 1998, Win98SP1 = 1999-2182, Win98SE >= 2183)

OSVersion PROC NEAR
OSVERSIONINFO STRUCT
dwOSVersionInfoSize   DWORD ?
dwMajorVersion        DWORD ?
dwMinorVersion        DWORD ?
dwBuildNumber         DWORD ?
dwPlatformId          DWORD ?
szCSDVersion          BYTE  128 dup (?)
OSVERSIONINFO ENDS

LOCALSPACE = (SIZE OSVERSIONINFO + 128) and -4  ; make space divisible by 4
        sub     esp, LOCALSPACE        ; reserve space for OSVERSIONINFO 
        mov     [esp].OSVERSIONINFO.dwOSVersionInfoSize, SIZE OSVERSIONINFO 
        INVOKE  GetVersionExA, esp
        test    eax, eax
        jz      OSV_Error              ; return 0 on error
        mov     eax, [esp].OSVERSIONINFO.dwMajorVersion
        shl     eax, 24
        mov     edx, [esp].OSVERSIONINFO.dwMinorVersion
        and     edx, 0FFH
        shl     edx, 16
        or      eax, edx
        mov     edx, [esp].OSVERSIONINFO.dwBuildNumber
        and     edx, 0000FFFFH
        or      eax, edx        
OSV_Error:
        add     esp, LOCALSPACE        ; free space
        ret
OSVersion ENDP

_TEXT   ENDS
        END     _start                 ; end of source code

